C*******************************************************************************
C
C                      IWFM STRATIGRAPHY DATA FILE
C
C                         Generated by pyiwfm
C                         {{ generation_time }}
C
C*******************************************************************************
C
C   This file defines the vertical stratigraphy (layer elevations) at each node.
C
C   NLAYERS ; Number of aquifer layers
C
C   FACTEL  ; Conversion factor for elevations (e.g., ft to meters)
C           ; Internal elevations = File elevations * FACTEL
C
C   For each node:
C     ID      ; Node identification number
C     GSELEV  ; Ground surface elevation
C     TOPELEV ; Top elevation of each layer (starting from layer 1)
C     BOTELEV ; Bottom elevation of each layer
C
C   Note: For layer i, the top elevation equals the bottom of layer i-1
C         (or ground surface for layer 1). Only bottom elevations are stored
C         per layer, with ground surface at the beginning.
C
C-----------------------------------------------------------------------------
{{ n_layers }}                                               / NLAYERS
{{ elev_factor | default(1.0) | fortran_float(14, 6) }}                                  / FACTEL
C
C   ID        GSELEV{% for i in range(n_layers) %}     BOT_L{{ i+1 }}{% endfor %}

C-----------------------------------------------------------------------------
