#{{ version }}
C***  DO NOT DELETE ABOVE LINE ***
C
C*******************************************************************************
C
C                  INTEGRATED WATER FLOW MODEL (IWFM)
C
C*******************************************************************************
C
C                    ROOT ZONE PARAMETERS DATA FILE
C                         Root Zone Component
C                         *** Version {{ version }} ***
C
C             Generated by pyiwfm
C             {{ generation_time }}
C
C*******************************************************************************
C                Root Zone Simulation Scheme Control and File Names
C
          {{ convergence }}                                      / RZCONV
          {{ max_iterations }}                                        / RZITERMX
          {{ inch_to_length }}                                  / FACTCN  (in -> ft)
{% if is_v411 %}
          {{ gw_uptake }}                                          / GWUPTK
{% endif %}
          {{ nonponded_file | pad_right(45) }}/ AGNPFL
          {{ ponded_file | pad_right(45) }}/ PFL
          {{ urban_file | pad_right(45) }}/ URBFL
          {{ native_riparian_file | pad_right(45) }}/ NVRVFL
          {{ return_flow_file | pad_right(45) }}/ RFFL
          {{ reuse_file | pad_right(45) }}/ RUFL
          {{ irig_period_file | pad_right(45) }}/ IPFL
          {{ moisture_src_file | pad_right(45) }}/ MSRCFL
          {{ ag_water_demand_file | pad_right(45) }}/ AGWDFL
          {{ lwu_budget_file | pad_right(45) }}/ LWUBUDFL
          {{ rz_budget_file | pad_right(45) }}/ RZBUDFL
{% if is_v411 %}
          {{ lwu_zbudget_file | pad_right(45) }}/ ZLWUBUDFL
          {{ rz_zbudget_file | pad_right(45) }}/ ZRZBUDFL
{% endif %}
{% if is_v412 %}
          {{ area_scale_file | pad_right(45) }}/ ARSCLFL
{% endif %}
{% if not is_v412 %}
          {{ final_moisture_file | pad_right(45) }}/ FMFL
{% endif %}
C*******************************************************************************
C             Parameters for Soil, Precipitation and Runoff Destination
C
          {{ k_factor }}                                    / FACTK   (cm -> ft)
{% if is_v41 %}
          {{ cprise_factor }}                                        / FACTEXDTH
{% endif %}
          {{ k_time_unit | pad_right(42) }}/ TUNITK
{% if is_v412 %}
          {{ dest_file | pad_right(45) }}/ DESTFL
{% endif %}
{{ column_header }}
