C  Wells definition file
C  Generated by pyiwfm
C
C  IWFM Groundwater Well Specification
C
C  ID      X              Y       ELEM  TOP_SCR  BOT_SCR  MAX_RATE  NAME
C
{{ n_wells }}                              / NWELLS
{% for well in wells %}
{{ well.id|fortran_int(6) }} {{ well.x|fortran_float(14, 4) }} {{ well.y|fortran_float(14, 4) }} {{ well.element|fortran_int(5) }} {{ well.top_screen|fortran_float(8, 2) }} {{ well.bottom_screen|fortran_float(8, 2) }} {{ well.max_pump_rate|fortran_float(10, 2) }}  {{ well.name }}
{% endfor %}
