C  Aquifer parameters file
C  Generated by pyiwfm
C
C  Hydraulic properties for {{ n_layers }} layer(s)
C
{{ n_nodes }}                              / NNODES
{{ n_layers }}                              / NLAYERS
C
C  NODE{% for i in range(n_layers) %}  KH{{ "%02d"|format(i+1) }}       KV{{ "%02d"|format(i+1) }}       SS{{ "%02d"|format(i+1) }}       SY{{ "%02d"|format(i+1) }}{% endfor %}

