/** Human-readable labels for budget type keys. */
export const BUDGET_LABELS: Record<string, string> = {
  gw: 'Groundwater',
  stream: 'Stream',
  stream_node: 'Stream Node',
  lwu: 'Land & Water Use',
  rootzone: 'Root Zone',
  unsaturated: 'Unsaturated Zone',
  diversion: 'Diversion',
  lake: 'Lake',
  small_watershed: 'Small Watershed',
};
