C*******************************************************************************
C                  GROUNDWATER BOUNDARY CONDITIONS MAIN FILE
C
C             Generated by pyiwfm
C             {{ generation_time }}
C*******************************************************************************
C
C   File Paths:
C     1. Specified Flow BC file (or blank)
C     2. Specified Head BC file (or blank)
C     3. General Head BC file (or blank)
C     4. Constrained General Head BC file (or blank)
C     5. Time Series BC data file (or blank)
C
C-------------------------------------------------------------------------------
          {{ spec_flow_bc_file | pad_right(45) }}/ SFBCFL
          {{ spec_head_bc_file | pad_right(45) }}/ SHBCFL
          {{ gen_head_bc_file | pad_right(45) }}/ GHBCFL
          {{ constrained_gh_bc_file | pad_right(45) }}/ CGHBCFL
          {{ ts_data_file | pad_right(45) }}/ TSBCDATAFL
C-------------------------------------------------------------------------------
C   NOUTB  ; Number of boundary node output locations
C-------------------------------------------------------------------------------
    {{ n_bc_output_nodes | fortran_int(10) }}                         / NOUTB
{% if n_bc_output_nodes > 0 %}
    {{ bc_output_file | pad_right(40) }} / BHYDOUTFL
{% for node in bc_output_nodes %}
     {{ node.id | fortran_int(5) }}      {{ node.layer | fortran_int(1) }}            {{ node.node | fortran_int(5) }}       {{ node.name }}
{% endfor %}
{% endif %}
