C  Stream nodes file
C  Generated by pyiwfm
C
C  IWFM Stream Node Specification
C
C  ID       X              Y       REACH  GW_NODE  BOT_ELEV  WET_PERM  UP_NODE  DN_NODE
C
{{ n_nodes }}                              / NSTRNODES
{% for node in nodes %}
{{ node.id|fortran_int(6) }} {{ node.x|fortran_float(14, 4) }} {{ node.y|fortran_float(14, 4) }} {{ node.reach_id|fortran_int(5) }} {{ (node.gw_node or 0)|fortran_int(7) }} {{ node.bottom_elev|fortran_float(10, 2) }} {{ node.wetted_perimeter|fortran_float(8, 2) }} {{ (node.upstream_node or 0)|fortran_int(7) }} {{ (node.downstream_node or 0)|fortran_int(7) }}
{% endfor %}
