C  Lake definitions file
C  Generated by pyiwfm
C
C  IWFM Lake Specification
C
C  ID   MAX_ELEV     INIT_STORAGE  NAME
C
{{ n_lakes }}                              / NLAKES
{% for lake in lakes %}
{{ lake.id|fortran_int(6) }} {{ lake.max_elevation|fortran_float(12, 4) }} {{ lake.initial_storage|fortran_float(14, 4) }}  {{ lake.name }}
{% endfor %}
