C*******************************************************************************
C
C                      IWFM NODAL COORDINATES FILE
C
C                         Generated by pyiwfm
C                         {{ generation_time }}
C
C*******************************************************************************
C
C   This file defines the X-Y coordinates of each node in the model.
C
C   NNODES  ; Number of nodes in the model domain
C
C   FACTXY  ; Conversion factor for nodal coordinates (e.g., ft to meters)
C           ; Internal coordinates = File coordinates * FACTXY
C
C   For each node:
C     ID    ; Node identification number (must be unique)
C     X     ; X-coordinate
C     Y     ; Y-coordinate
C
C-----------------------------------------------------------------------------
{{ n_nodes }}                                               / NNODES
{{ coord_factor | default(1.0) | fortran_float(14, 6) }}                                  / FACTXY
C
C   ID              X              Y
C-----------------------------------------------------------------------------
