C*******************************************************************************
C
C                      IWFM ELEMENT CONFIGURATION FILE
C
C                         Generated by pyiwfm
C                         {{ generation_time }}
C
C*******************************************************************************
C
C   This file defines the element connectivity and subregion assignments.
C
C   NELEM      ; Number of elements in the model domain
C   NSUBREGION ; Number of subregions
C
C   Subregion names follow (one per line)
C
C   For each element:
C     ID    ; Element identification number (must be unique)
C     V1-V4 ; Node IDs of vertices (counterclockwise; 0 for 4th if triangle)
C     SR    ; Subregion number (1 to NSUBREGION)
C
C-----------------------------------------------------------------------------
{{ n_elements }}                                            / NELEM
{{ n_subregions }}                                            / NSUBREGION
C
C   Subregion Names (ID # Name)
C-----------------------------------------------------------------------------
{% for sr_id, sr_name in subregions %}
{{ sr_id }} / {{ sr_name }}
{% endfor %}
C
C   ID     V1     V2     V3     V4   SR
C-----------------------------------------------------------------------------
